/*
 *  $Id: tracer.h 707 2008-03-08 10:06:44Z bkr $
 *
 *  Copyright (c) 2007 Bjoern Kriews
 */

#ifndef __TRACER_H__
#define __TRACER_H__ 1

#include "er_util.h"

#ifndef TRACER_DEBUG
#define TRACER_DEBUG 0
#endif

#ifndef TRACER_DEBUG_FAT
#if TRACER_DEBUG
#define TRACER_DEBUG_FAT 1
#else
#define TRACER_DEBUG_FAT 0
#endif
#endif

#ifndef TRACER_DEBUG_MACHO
#if TRACER_DEBUG
#define TRACER_DEBUG_MACHO 1
#else
#define TRACER_DEBUG_MACHO 0
#endif
#endif

#ifndef TRACER_DEBUG_SYM
#if TRACER_DEBUG
#define TRACER_DEBUG_SYM 1
#else
#define TRACER_DEBUG_SYM 0
#endif
#endif

#ifndef TRACER_DEBUG_SYM_HEAVY
#define TRACER_DEBUG_SYM_HEAVY 0
#endif

#ifndef TRACER_DEBUG_STORE
#if TRACER_DEBUG
#define TRACER_DEBUG_STORE 1
#else
#define TRACER_DEBUG_STORE 0
#endif
#endif

#ifndef TRACER_DEBUG_LOOKUP
#if TRACER_DEBUG
#define TRACER_DEBUG_LOOKUP 1
#else
#define TRACER_DEBUG_LOOKUP 0
#endif
#endif

/* list of symbols to restrict heavy debug output to - i.e.: */
/* #define TRACER_DEBUG_LIMIT "_start _main __nsnote_callback _tracerDumpBacktrace" */

#define TRACER_SUCCESS 0
#define TRACER_FAILURE 1

#define TRACER_OPT_DEFAULT 0

typedef struct tracerSymbol {
  uint32_t strtab_index;
  uint32_t strtab_imagePath_index;
  uint32_t strtab_imageName_index;
  uint64_t value;

} TRACER_SYMBOL;


typedef struct tracerSymbolStore {
  TRACER_SYMBOL *symbols;
  uint32_t symbols_size;
  uint32_t symbols_capacity;
  uint32_t symbols_capacity_growth;

  const char *strtab;
  uint32_t strtab_size;
  uint32_t strtab_capacity;
  uint32_t strtab_capacity_growth;

  /* used internally while collecting */
  const char *current_imagePath;
  uint32_t current_strtab_imageName_index;
  uint32_t current_strtab_imagePath_index;

} TRACER_SYMBOL_STORE;

typedef int (*TRACER_SYMBOL_WORKER)(const char *imagePath, const char *symbolName, uint64_t symbolAddress, void *workerInfo, uint64_t options);
typedef int (*TRACER_SYMBOL_SORTER)(void *sorterInfo, const TRACER_SYMBOL *a, const TRACER_SYMBOL *b);

int tracerAnalyzeFile(UTIL_MACH_HEADER *mhp, const char *imageName, intptr_t slide, TRACER_SYMBOL_WORKER symbolAction, void *symbolActionInfo, uint64_t options);
int tracerAnalyzeMachO(UTIL_MACH_HEADER *mhp, const char *imageName, UTIL_MACH_HEADER *live_mhp, TRACER_SYMBOL_WORKER symbolAction, void *symbolActionInfo, uint64_t options );
int tracerWalkSymbolTables(TRACER_SYMBOL_WORKER symbolAction, void *symbolActionInfo, uint64_t options );


/*
 * symbol store management
 */
int tracerSymbolStoreCollector(const char *imageName, const char *name, uint64_t value, void *info, uint64_t options);

TRACER_SYMBOL_STORE *tracerSymbolStoreCollect(uint64_t options, TRACER_SYMBOL_SORTER sorter, void *sorterInfo);

TRACER_SYMBOL_STORE *tracerSymbolStoreForLookupByAddress(void);

int tracerSymbolStoreInit( TRACER_SYMBOL_STORE *store );
uint32_t tracerSymbolStoreCount( TRACER_SYMBOL_STORE *store );

const char *tracerSymbolStoreNameOfSymbol( TRACER_SYMBOL_STORE *store, TRACER_SYMBOL *symbol );
const char *tracerSymbolStoreImageNameOfSymbol( TRACER_SYMBOL_STORE *store, TRACER_SYMBOL *symbol );
const char *tracerSymbolStoreImagePathOfSymbol( TRACER_SYMBOL_STORE *store, TRACER_SYMBOL *symbol );

int tracerSymbolStoreShow( TRACER_SYMBOL_STORE *store );
int tracerSymbolStoreShowContents( TRACER_SYMBOL_STORE *store );

int tracerSymbolStoreSorterByAddress( void *sorterInfo, const TRACER_SYMBOL *a, const TRACER_SYMBOL *b );
int tracerSymbolStoreSort( TRACER_SYMBOL_STORE *store, TRACER_SYMBOL_SORTER sorter, void *sorterInfo );

int tracerSymbolStoreLookupByAddress( TRACER_SYMBOL_STORE *store, uint64_t address, TRACER_SYMBOL **symbolPtr, uint64_t *offsetPtr );

TRACER_SYMBOL_STORE *tracerSymbolStoreAllocate(void);
void tracerSymbolStoreRelease( TRACER_SYMBOL_STORE *store );

void _tracerDump(int skipframes, void **pcs, int frames);

#define tracerDump() \
  ({ void *__pcs[64]; _tracerDump( 0 /*skipframes*/, __pcs, backtrace(__pcs, sizeof(__pcs)/sizeof(*__pcs))); })

int tracerCTest(int argc, const char **argv);


/*
 * support for gcc -finstrument-functions
 */

typedef struct tracerEvent {
  unsigned long eventType;
  void *function;
  void *call_site;
  void *thread;
} TRACER_EVENT;


#define TRACER_EVENT_VERSION 1

#define TRACER_EVENT_TYPE_MAGIC_32 0xdecafbadUL
#ifdef UTIL_64BIT
#define TRACER_EVENT_TYPE_MAGIC_64 0xbadcafe0babedeadUL
#define TRACER_EVENT_TYPE_BEGIN TRACER_EVENT_TYPE_MAGIC_64
#else
#define TRACER_EVENT_TYPE_MAGIC_64 0xbadcafe0babedeadULL
#define TRACER_EVENT_TYPE_BEGIN TRACER_EVENT_TYPE_MAGIC_32
#endif

#define TRACER_EVENT_TYPE_END   'E'
#define TRACER_EVENT_TYPE_ENTER 'I'
#define TRACER_EVENT_TYPE_EXIT  'O'

void tracerStartInstrumentation(void);
void tracerStopInstrumentation(void);
void tracerInstrumentationEvent(unsigned long eventType, void *this_fn, void *call_site);

#ifdef __OBJC__
#import <Foundation/Foundation.h>

typedef struct tracerRuntimeScannerInfo {
  TRACER_SYMBOL_STORE *symbolStore;

  uint32_t classCount;
  uint32_t methodCount;
        
  NSMutableArray *classArray;
  NSMutableArray *methodArray;
  NSMutableDictionary *classDictionary;
  NSArray *sortedIMPs;
        
  // internal 
  NSString *currentClassName;
  NSMutableArray *currentMethodArray;
        
  uint64_t searchAddress;
  uint64_t offset;
  uint64_t nearestSymbolAddress;
  NSString *nearestSymbol;
  Class nearestClass;
  SEL nearestSel;

  NSString *imagePath;

} TRACER_RUNTIME_SCANNER_INFO;

void tracerRuntimeScannerInfoInit(TRACER_RUNTIME_SCANNER_INFO *info);

void tracerScanMethodList(Class aClass, TRACER_RUNTIME_SCANNER_INFO *info);
BOOL tracerScanClassList(TRACER_RUNTIME_SCANNER_INFO *info);

// keys for frame info dictionaries returned by tracerFrameInfoForAddress
extern NSString *kTracerAddress;
extern NSString *kTracerSymbolOffset;
extern NSString *kTracerSymbolAddress;
extern NSString *kTracerSymbolName;
extern NSString *kTracerClassName;
extern NSString *kTracerSelectorName;
extern NSString *kTracerIsClassMethod;
extern NSString *kTracerImagePath;
extern NSString *kTracerImageName;

NSDictionary *tracerFrameInfoForAddress(uint64_t theAddress, TRACER_RUNTIME_SCANNER_INFO *info);

NSArray *_tracerBacktraceAsArrayOfDictionaries(int skipframes, void **pcs, int frames);
NSArray *_tracerBacktraceAsArrayOfStrings(NSArray *traceDictionaries);

#define tracerBacktraceAsArrayOfDictionaries(skipframes) \
({ void *__pcs[64]; _tracerBacktraceAsArrayOfDictionaries( skipframes, __pcs, backtrace(__pcs, sizeof(__pcs)/sizeof(*__pcs))); })

#define tracerBacktraceAsArrayOfStrings(skipframes) \
_tracerBacktraceAsArrayOfStrings( tracerBacktraceAsArrayOfDictionaries(skipframes) )

#define tracerBacktraceAsString(skipframes) \
( [[tracerBacktraceAsArrayOfStrings(skipframes) componentsJoinedByString: @"\n"] stringByAppendingString: @"\n"] )

#define tracerDumpBacktrace(skipframes) \
NSLog(@"\nBacktrace:\n%@\n", tracerBacktraceAsString(0) )


int tracerObjCTest(int argc, const char **argv);

#endif

#endif 
/* __TRACER_H__ */

