//
//  $Id: tracer.m 707 2008-03-08 10:06:44Z bkr $
//

#include "tracer.c"

#import <objc/runtime.h>

NSString *kTracerAddress       = @"kTracerSearchAddress";
NSString *kTracerSymbolOffset  = @"kTracerSymbolOffset";
NSString *kTracerSymbolAddress = @"kTracerSymbolAddress";
NSString *kTracerSymbolName    = @"kTracerSymbolName";
NSString *kTracerClassName     = @"kTracerClassName";
NSString *kTracerSelectorName  = @"kTracerSelectorName";
NSString *kTracerIsClassMethod = @"kTracerIsClassMethod";
NSString *kTracerImagePath     = @"kTracerImagePath";
NSString *kTracerImageName     = @"kTracerImageName";

void NO_INSTRUMENT
tracerRuntimeScannerInfoInit(TRACER_RUNTIME_SCANNER_INFO *info)
{
  memset(info, 0, sizeof(*info));
}

void NO_INSTRUMENT 
tracerScanMethodList(Class aClass, TRACER_RUNTIME_SCANNER_INFO *info)
{
	unsigned int methodCount;
	Method *theMethods = class_copyMethodList( aClass, &methodCount);
	
	NSMutableString *methodNamePrefix = nil;
	BOOL isMetaClass = NO;
	
	info->methodCount += methodCount;
	
	unsigned int methodIdx;
	for( methodIdx = 0 ; methodIdx < methodCount ; ++methodIdx ) {
		Method aMethod = theMethods[ methodIdx ];
		uint64_t imp = UTIL_PTR_AS_ULL( method_getImplementation( aMethod ) );
		
		if( info->searchAddress >= imp ) {
		  uint64_t impOffset = info->searchAddress - imp;
		  
		  tracerDebugSymHeavy("search: %08llx ioff: %08llx off: %08llx [%s %s]", info->searchAddress, (unsigned long long) info->offset, (unsigned long long) impOffset, class_getName( aClass ), sel_getName( method_getName( aMethod) ) );

		  if( impOffset < info->offset ) {
		    info->nearestSymbolAddress = imp; 
		    info->nearestClass = aClass;
		    info->nearestSel = method_getName( aMethod );
		    info->offset = impOffset;		    
		  }
		}

		if( info->currentMethodArray || info->methodArray ) {
			if ( nil == methodNamePrefix ) {
				isMetaClass = class_isMetaClass( aClass );
				methodNamePrefix = [NSMutableString stringWithString: isMetaClass ? @"+[" : @"-["];
				
				[methodNamePrefix appendString: info->currentClassName];
				[methodNamePrefix appendString: @" "];
			}
			
			NSString *selectorName = NSStringFromSelector( method_getName( aMethod ) );
			NSString *fullName = [[methodNamePrefix stringByAppendingString: selectorName] stringByAppendingString: @"]"];
			
			NSNumber *impNumber = [NSNumber numberWithUnsignedLongLong: (unsigned long long) imp];
			NSNumber *isClassMethod = [NSNumber numberWithBool: isMetaClass ? YES : NO];
				
			NSDictionary *methodDescription = [NSDictionary dictionaryWithObjectsAndKeys:
											   impNumber, @"symbolAddress",
											   info->currentClassName, @"className",
											   selectorName, @"selectorName",
											   fullName, @"symbolName",
											   isClassMethod, @"isClassMethod",
											   nil];
			
			if( info->currentMethodArray ) 
				[info->currentMethodArray addObject: methodDescription];
				
			if( info->methodArray ) 
				[info->methodArray addObject: methodDescription];
		}

	}

	free(theMethods);
}

BOOL NO_INSTRUMENT
tracerScanClassList(TRACER_RUNTIME_SCANNER_INFO *info)
{
  info->classCount = 0;
  info->methodCount = 0;
  info->currentMethodArray = nil;

  unsigned int newClassCount = objc_getClassList(NULL, 0);

    Class *theClasses = NULL;
    while (info->classCount < newClassCount) {
        info->classCount = newClassCount;
        theClasses = realloc(theClasses, sizeof(Class) * info->classCount);
        newClassCount = objc_getClassList(theClasses, info->classCount);
    }

    if( NULL == theClasses )
      return NO;

	unsigned int classIdx;
	for(classIdx = 0 ; classIdx < info->classCount ; ++classIdx ) {
		
		Class aClass = theClasses[ classIdx ];

		info->currentClassName = NSStringFromClass( aClass );

//		if(! [info->currentClassName hasPrefix: @"Little"] && ![info->currentClassName hasPrefix: @"NSButton"] ) 
//			continue;
			
		if( info->classArray ) {
			[info->classArray addObject: info->currentClassName];
		}
			
		if( info->classArray ) {
			info->currentMethodArray = [NSMutableArray array];
			[info->classDictionary setObject: info->currentMethodArray forKey: info->currentClassName];
		}
		
		Class aMetaClass = objc_getMetaClass( class_getName( aClass ) );
		
		tracerScanMethodList(aMetaClass, info);
		tracerScanMethodList(aClass, info);
	}
	
	free(theClasses);    

#if 0
	// sortUsingKeyPath is from a category
	if( info->methodArray ) {
		[info->methodArray sortUsingKeyPath: @"symbolAddress" ascending: YES];
		info->sortedIMPs = [info->methodArray valueForKey: @"symbolAddress"];

	}
#endif

	return YES;
}

NSDictionary * NO_INSTRUMENT
tracerFrameInfoForAddress(uint64_t theAddress, TRACER_RUNTIME_SCANNER_INFO *info)
{
	info->imagePath = nil;
	info->nearestSymbol = nil;
	info->nearestSymbolAddress = 0;
	info->nearestClass = Nil;
	info->nearestSel = NULL;
	info->offset = 0;

	info->searchAddress = theAddress;
	
	if( NULL != info->symbolStore ) {
	  TRACER_SYMBOL *functionSymbol = NULL;

	  if( TRACER_SUCCESS == tracerSymbolStoreLookupByAddress(info->symbolStore, info->searchAddress, &functionSymbol, &info->offset) ) {
	    info->nearestSymbol = [NSString stringWithUTF8String: tracerSymbolStoreNameOfSymbol(info->symbolStore, functionSymbol)];
	    info->imagePath = [NSString stringWithUTF8String: tracerSymbolStoreImagePathOfSymbol(info->symbolStore, functionSymbol)];
	    info->nearestSymbolAddress = functionSymbol->value;
	  }
	}
  
	// scans all methods, updating fields in the info structure is something nearer is found
	tracerScanClassList(info);

	NSMutableDictionary *frame = [[NSMutableDictionary alloc] initWithCapacity: 10];

	if( info->nearestClass ) {
	  
	  BOOL isClassMethod = class_isMetaClass( info->nearestClass );
	  NSString *className = [NSString stringWithUTF8String: class_getName( info->nearestClass )];
	  NSString *selectorName = [NSString stringWithUTF8String: sel_getName( info->nearestSel )];

	  info->nearestSymbol = [NSString stringWithFormat: @"%c[%@ %@]", isClassMethod ? '+' : '-', className, selectorName];

	  [frame setValue: [NSNumber numberWithBool: isClassMethod] forKey: kTracerIsClassMethod];
	  [frame setValue: className forKey: kTracerClassName];
	  [frame setValue: selectorName forKey: kTracerSelectorName];
	}

	[frame setValue: [NSNumber numberWithUnsignedLongLong: (unsigned long long) info->searchAddress] forKey: kTracerAddress];
	[frame setValue: [NSNumber numberWithUnsignedLongLong: (unsigned long long) info->offset] forKey: kTracerSymbolOffset];
	[frame setValue: [NSNumber numberWithUnsignedLongLong: (unsigned long long) info->nearestSymbolAddress] forKey: kTracerSymbolAddress];
	[frame setValue: info->nearestSymbol forKey: kTracerSymbolName];

	[frame setValue: info->imagePath forKey: kTracerImagePath];
	[frame setValue: [info->imagePath lastPathComponent] forKey: kTracerImageName];

	tracerDebugSym("addr: 0x%llx image: %s sym: %s (0x%llx) + %lld\n",
	      (unsigned long long)info->searchAddress, [info->imagePath UTF8String], [info->nearestSymbol UTF8String], (unsigned long long)info->nearestSymbolAddress, (unsigned long long) info->offset);	      
	

	NSDictionary *ret = [[frame mutableCopy] autorelease];
	[frame release];

	tracerDebugSym("frameInfo: %s", [[ret description] UTF8String]);

	return ret;
}

#if 0
void NO_INSTRUMENT
tracerUnused(void)
{
	info.classArray = [NSMutableArray array];
	info.classDictionary = [NSMutableDictionary dictionary];

	info.symbolStore = tracerSymbolStoreForLookupByAddress();

	info.methodArray = [NSMutableArray array];
}
#endif

NSArray * NO_INSTRUMENT
_tracerBacktraceAsArrayOfDictionaries(int skipFrames, void **pcs, int frames)
{
	TRACER_RUNTIME_SCANNER_INFO info;
	tracerRuntimeScannerInfoInit(&info);

	info.symbolStore = tracerSymbolStoreForLookupByAddress();
			
	NSMutableArray *trace = [NSMutableArray array];

	int i;
	for (i = skipFrames ; i < frames ; ++i) { 
	  if( pcs[i] > (void*) 0x8 ) /* there is a bug where argc shows up as a first frame if it is in the range 1..3 (OS X 10.5.2) */
	    [trace addObject: tracerFrameInfoForAddress( UTIL_PTR_AS_ULL( pcs[i] ), &info)];
	}

	tracerSymbolStoreRelease( info.symbolStore );

	return trace;
}

NSArray * NO_INSTRUMENT
_tracerBacktraceAsArrayOfStrings(NSArray *trace)
{
  int frames = [trace count];
  
  NSMutableArray *strings = [NSMutableArray arrayWithCapacity: frames];
  
  int i;
  for( i = 0 ; i < frames ; ++i ) {
    NSDictionary *frameInfo = [trace objectAtIndex: i];
    
    [strings addObject: 
#ifdef __LP64__		 
	       [NSString stringWithFormat: @"%-3d %-35.35s 0x%016llx %@ + %lld", // format like backtrace_symbols()
#else
			 [NSString stringWithFormat: @"%-3d %-35.35s 0x%08llx %@ + %lld",
#endif
				   i,
				   [[frameInfo valueForKey: kTracerImageName] UTF8String],
				   [[frameInfo valueForKey: kTracerAddress] unsignedLongLongValue],
				   [frameInfo valueForKey: kTracerSymbolName],
				   [[frameInfo valueForKey: kTracerSymbolOffset] unsignedLongLongValue]
			  ]];
	     }
      
      return strings;
  }


#ifdef TRACER_TEST

@interface TracerTestObject : NSObject
- (void) tracerTestMethod: (id) someArg;
+ (void) dumpBacktrace;
@end

@implementation TracerTestObject

- (void) tracerTestMethod: (id) someArg
{
  [[self class] dumpBacktrace];
}

+ (void) dumpBacktrace
{
  // for comparing 
  // tracerDump();

  tracerDumpBacktrace();
}
@end

void tracerObjCDummy(TracerTestObject *t)
{
  fprintf(stderr, "Direct\n");

  [t tracerTestMethod: @"TestArg"];

  fprintf(stderr, "Through NSArray\n");

  NSArray *a = [NSArray arrayWithObject: t];

  [a makeObjectsPerformSelector: @selector(tracerTestMethod:) withObject: @"TestArg2"];
}

int tracerObjCTest(int argc, const char **argv)
{
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  tracerBeginInstrumentation();

  TracerTestObject *test = [[TracerTestObject alloc] init];
  
  // IMP imp = [NSArray instanceMethodForSelector: @selector(makeObjectsPerformSelector:withObject:)];
  // fprintf(stderr, "imp: %p\n", imp);
  
  tracerObjCDummy(test);
  
  [test release];

  tracerEndInstrumentation();
    
  [pool release];

  return 0;
}

int main(int argc, const char **argv)
{
  return tracerObjCTest(argc, argv);
}

#endif
/* TRACER_TEST */
